(* Supporting data types for indexing *)
(* Simple version *)
(* Author: Frank Pfenning *)

(* Ordered Keys *)
signature ORDER =
sig
  type key
  val compare : key * key -> order
end;

(* Integers *)
structure IntOrder : ORDER where type key = int =
struct
  type key = int
  val compare = Int.compare
end;

(* Ordered Sets *)
signature SET =
sig
  structure O : ORDER

  type set

  val empty : set
  val insert : O.key * set -> set

  val union : set * set -> set
  val intersect : set * set -> set

  val size : set -> int
end;

(* 
   Ordered sets as ordered lists
   Some alternative implementations: dictionaries, bit vectors
*)
functor Set
  (structure Order : ORDER)
  : SET where type O.key = Order.key =
struct
  structure O = Order

  type set = O.key list

  val empty = nil
  fun insert (x, set) =
      let
	fun ins nil = x::nil
	  | ins (l as y::l') =
	    (case O.compare (x, y)
	       of EQUAL => l
	        | LESS => x::l
		| GREATER => y::ins(l'))
      in
	ins (set)
      end
 
  fun union (nil, k) = k
    | union (l, nil) = l
    | union (l as x::l', k as y::k') =
      (case O.compare (x, y)
         of EQUAL => x::union(l', k')
          | LESS => x::union(l', k)
          | GREATER => y::union(l, k'))

  fun intersect (nil, k) = nil
    | intersect (l, nil) = nil
    | intersect (l as x::l', k as y::k') =
      (case O.compare (x, y)
         of EQUAL => x::intersect(l', k')
          | LESS => intersect(l', k)
          | GREATER => intersect(l, k'))

  fun size (l) = List.length (l)
end;

(* Sets of integers *)
structure IntSet : SET where type O.key = int =
  Set (structure Order = IntOrder);
